"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Expression = exports.NCalcErrorListener = void 0;
const antlr4_1 = require("antlr4");
const Domain_1 = require("./Domain");
const EvaluationOptions_1 = require("./EvaluationOptions");
const NCalcLexer_1 = __importDefault(require("../Grammar/NCalcLexer"));
const NCalcParser_1 = __importDefault(require("../Grammar/NCalcParser"));
class NCalcErrorListener {
    _errors = [];
    get errors() {
        return this._errors;
    }
    syntaxError(...args) {
        this._errors.push(args);
    }
}
exports.NCalcErrorListener = NCalcErrorListener;
class Expression {
    lexerErrors;
    parserErrors;
    Options = EvaluationOptions_1.EvaluateOptions.None;
    /**
     * Orginal strings representation of the expression
     */
    OriginalExpression;
    static _cacheEnabled = true;
    static _compiledExpression = {};
    ParsedExpression;
    ParameterEnumerators = {};
    ParametersBackup = {};
    Parameters = {};
    get CacheEnabled() {
        return Expression._cacheEnabled;
    }
    static get CachedExpressions() {
        return Expression._compiledExpression;
    }
    set CacheEnabled(value) {
        Expression._cacheEnabled = value;
        if (value === false) {
            Expression._compiledExpression = {};
        }
    }
    constructor(expression, options = EvaluationOptions_1.EvaluateOptions.None) {
        if (expression == null || expression == '') {
            throw new Error('The expression cannot be null or empty');
        }
        if (typeof expression === 'string') {
            this.OriginalExpression = expression;
        }
        else {
            this.ParsedExpression = expression;
        }
        if (options) {
            this.Options = options;
        }
        this.lexerErrors = new NCalcErrorListener();
        this.parserErrors = new NCalcErrorListener();
    }
    get errors() {
        return this.lexerErrors.errors.concat(this.parserErrors.errors);
    }
    Compile(expression, nocache) {
        let logicalExpression = null;
        if (this.CacheEnabled && !nocache) {
            if (Expression._compiledExpression.hasOwnProperty(expression)) {
                const wr = Expression._compiledExpression[expression];
                const stored = wr.deref();
                if (stored && stored !== undefined) {
                    return stored;
                }
            }
        }
        if (logicalExpression == null) {
            // Create the lexer
            let inputStream = new antlr4_1.CharStream(expression);
            let lexer = new NCalcLexer_1.default(inputStream);
            lexer.addErrorListener(this.lexerErrors);
            // Create parser
            let tokenStream = new antlr4_1.CommonTokenStream(lexer);
            let parser = new NCalcParser_1.default(tokenStream);
            parser.addErrorListener(this.parserErrors);
            logicalExpression = parser.GetExpression();
            if (this.CacheEnabled && !nocache) {
                Expression._compiledExpression[expression] = new WeakRef(logicalExpression);
            }
        }
        return logicalExpression;
    }
    /**
     * Detects whether the expression has errors. This will simply return a boolean value.
     * You can access the error by using the `errors` getter.
     */
    HasErrors() {
        try {
            if (this.ParsedExpression == null) {
                this.ParsedExpression = this.Compile(this.OriginalExpression, (this.Options & EvaluationOptions_1.EvaluateOptions.NoCache) == EvaluationOptions_1.EvaluateOptions.NoCache);
            }
            if (this.errors.length > 0) {
                throw new Error();
            }
            // In case HasErrors() is called multiple times for the same expression
            return this.ParsedExpression === null || this.ParsedExpression === undefined;
        }
        catch (e) {
            return true;
        }
    }
    EvaluateFunction = {};
    EvaluateParameter = {};
    Evaluate() {
        if (this.HasErrors()) {
            throw new Domain_1.EvaluationException('Failed evaluating the expression. Refer to errors.');
        }
        if (this.ParsedExpression == null) {
            this.ParsedExpression = this.Compile(this.OriginalExpression, (this.Options & EvaluationOptions_1.EvaluateOptions.NoCache) == EvaluationOptions_1.EvaluateOptions.NoCache);
        }
        var visitor = new Domain_1.EvaluationVisitor(this.Options);
        visitor.EvaluateFunction = this.EvaluateFunction;
        visitor.EvaluateParameter = this.EvaluateParameter;
        visitor.Parameters = this.Parameters;
        // if array evaluation, execute the same expression multiple times
        if ((this.Options & EvaluationOptions_1.EvaluateOptions.IterateParameters) == EvaluationOptions_1.EvaluateOptions.IterateParameters) {
            let size = -1;
            this.ParametersBackup = {};
            for (let key in this.Parameters) {
                this.ParametersBackup[key] = this.Parameters[key];
            }
            this.ParameterEnumerators = {};
            for (let parameter in this.Parameters) {
                const value = this.Parameters[parameter];
                if (Array.isArray(value)) {
                    let localsize = value.length;
                    if (size == -1) {
                        size = localsize;
                    }
                    else if (localsize != size) {
                        throw new Domain_1.EvaluationException('When IterateParameters option is used, IEnumerable parameters must have the same number of items');
                    }
                }
            }
            for (let key in this.Parameters) {
                var parameter = this.Parameters[key];
                if (parameter != null) {
                    this.ParameterEnumerators[key] = parameter;
                }
            }
            var results = [];
            for (let i = 0; i < size; i++) {
                for (let key in this.ParameterEnumerators) {
                    let enumerator = this.ParameterEnumerators[key];
                    this.Parameters[key] = enumerator[i];
                }
                this.ParsedExpression.Accept(visitor);
                results.push(visitor.Result);
            }
            return results;
        }
        this.ParsedExpression.Accept(visitor);
        return visitor.Result;
    }
}
exports.Expression = Expression;
